#include "SPPi.h"
// To enable serial debugging see "settings.h"
#define EXTRADEBUG // Uncomment to get even more debugging data
#define PRINTREPORT // Uncomment to print the report sent to the Arduino


SPPi::SPPi(BTD *p, const char* name, const char* pin, bool pair, uint8_t *addr) :
pBtd(p) // Pointer to BTD class instance - mandatory
{
        if (pBtd)
                pBtd->registerServiceClass(this); // Register it as a Bluetooth service

        pBtd->btdName = name;
        pBtd->btdPin = pin;
        pBtd->pairWithOtherDevice = pair;
        
        for (uint8_t i = 0; i < 6; i++)
                pBtd->remote_bdaddr[i] = addr[i];

        /* Set device cid for the SDP and RFCOMM channelse */
        sdp_scid[0] = 0x41; // 0x0050
        sdp_scid[1] = 0x00;

        rfcomm_scid[0] = 0x51; // 0x0051
        rfcomm_scid[1] = 0x00;

        Reset();
}

void SPPi::Reset() {
        connected = false;
        RFCOMMConnected = false;
        SDPConnected = false;
        waitForLastCommand = false;
        l2cap_sdp_state = L2CAP_SDP_WAIT;
        l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
        l2cap_event_flag = 0;
        sppIndex = 0;
}

void SPPi::disconnect() {
        connected = false;
        // First the two L2CAP channels has to be disconencted and then the HCI connection
        if (RFCOMMConnected)
                pBtd->l2cap_disconnection_request(hci_handle, 0x0A, rfcomm_scid, rfcomm_dcid);
        if (RFCOMMConnected && SDPConnected)
                delay(1); // Add delay between commands
        if (SDPConnected)
                pBtd->l2cap_disconnection_request(hci_handle, 0x0B, sdp_scid, sdp_dcid);
        l2cap_sdp_state = L2CAP_DISCONNECT_RESPONSE;
}

void SPPi::ACLData(uint8_t* l2capinbuf) {

/*        Notify(PSTR("\r\n--- Incoming Packet: "), 0x80);

        for (uint8_t i = 0; i < (l2capinbuf[2] + 4); i++) {
                D_PrintHex<uint8_t > (l2capinbuf[i], 0x80);
                Notify(PSTR(" "), 0x80);
        }
        Notify(PSTR("\r\n---"), 0x80);*/


        if (((l2capinbuf[0] | (l2capinbuf[1] << 8)) == (hci_handle | 0x2000))) { // acl_handle_ok
//                Notify(PSTR("\r\nacl_handle_ok"), 0x80);
                if ((l2capinbuf[6] | (l2capinbuf[7] << 8)) == 0x0001) { //l2cap_control - Channel ID for ACL-U

                        if (l2capinbuf[8] == L2CAP_CMD_COMMAND_REJECT) {
                                Notify(PSTR("\r\nL2CAP Command Rejected - Reason: "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[13], 0x80);
                                Notify(PSTR(" "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[12], 0x80);
                                Notify(PSTR(" Data: "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[17], 0x80);
                                Notify(PSTR(" "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[16], 0x80);
                                Notify(PSTR(" "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[15], 0x80);
                                Notify(PSTR(" "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[14], 0x80);
                        }else if (l2capinbuf[8] == L2CAP_CMD_CONNECTION_RESPONSE) {
                                if (((l2capinbuf[16] | (l2capinbuf[17] << 8)) == 0x0000) && ((l2capinbuf[18] | (l2capinbuf[19] << 8)) == SUCCESSFUL)) { // Success
                                        Notify(PSTR("\r\nL2CAP Connection Response "), 0x80);
                                        if (l2capinbuf[14] == sdp_scid[0] && l2capinbuf[15] == sdp_scid[1]) {
                                                Notify(PSTR("\r\nSDP Connection Response"), 0x80);
                                                //identifier = l2capinbuf[9];
                                                sdp_dcid[0] = l2capinbuf[12];
                                                sdp_dcid[1] = l2capinbuf[13];

                                                //Notify(PSTR("\r\ndcid: "), 0x80);
                                                //D_PrintHex<uint8_t > (l2capinbuf[12], 0x80);
                                                //Notify(PSTR(" "), 0x80);
                                                //D_PrintHex<uint8_t > (l2capinbuf[13], 0x80);

                                                l2cap_event_flag |= L2CAP_FLAG_CONNECTION_SDP_RESPONSE;
                                        }
                                } 

                        } else if(l2capinbuf[8] == L2CAP_CMD_CONFIG_RESPONSE) {
                                if((l2capinbuf[16] | (l2capinbuf[17] << 8)) == 0x0000) { // Success
                                        if(l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {
                                                Notify(PSTR("\r\nSDP Configuration Response Received"), 0x80);
                                                l2cap_event_flag |= L2CAP_FLAG_CONFIG_SDP_SUCCESS;
                                        }
                                }

                        } else if (l2capinbuf[8] == L2CAP_CMD_CONFIG_REQUEST) {
                                if (l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {

                                        Notify(PSTR("\r\nSDP Configuration Request Received"), 0x80);

                                        identifier = l2capinbuf[9];
                                        l2cap_event_flag |= L2CAP_FLAG_CONFIG_SDP_REQUEST;
                                }
                        } else if (l2capinbuf[8] == L2CAP_CMD_DISCONNECT_REQUEST) {
                                if (l2capinbuf[12] == sdp_scid[0] && l2capinbuf[13] == sdp_scid[1]) {
                                        Notify(PSTR("\r\nDisconnect Request: SDP Channel"), 0x80);
                                        identifier = l2capinbuf[9];
                                        l2cap_event_flag |= L2CAP_FLAG_DISCONNECT_SDP_REQUEST;
                                }
                        } else if (l2capinbuf[8] == L2CAP_CMD_DISCONNECT_RESPONSE) {
                                if (l2capinbuf[12] == sdp_dcid[0] && l2capinbuf[13] == sdp_dcid[1]) {
                                        Notify(PSTR("\r\nDisconnect Response: SDP Channel"), 0x80);
                                        //identifier = l2capinbuf[9];
                                        l2cap_event_flag |= L2CAP_FLAG_DISCONNECT_RESPONSE;
                                }
                        } else if (l2capinbuf[8] == L2CAP_CMD_INFORMATION_REQUEST) {
                                Notify(PSTR("\r\nInformation request"), 0x80);
                                identifier = l2capinbuf[9];
                                pBtd->l2cap_information_response(hci_handle, identifier, l2capinbuf[12], l2capinbuf[13]);
                        }
                        else {
                                Notify(PSTR("\r\nL2CAP Unknown Signaling Command: "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[8], 0x80);
                        }
                } else if (l2capinbuf[6] == sdp_scid[0] && l2capinbuf[7] == sdp_scid[1]) { // SDP

                        Notify(PSTR("\r\nSDP data..."), 0x80);

                        if (l2capinbuf[8] == SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU) {
                                l2cap_event_flag |= L2CAP_FLAG_SERVICE_SEARCH_ATTR_RESPONSE;
                                if (l2cap_sdp_state == L2CAP_SDP_SERVICE_SEARCH_ATTR1)
                                { 
                                        Notify(PSTR(" - SDP Service Search Attribute Response 1: "), 0x80);
                                        remainingBytes = l2capinbuf[l2capinbuf[2] + 3];

                                }else if (l2cap_sdp_state == L2CAP_SDP_SERVICE_SEARCH_ATTR2){
                                        Notify(PSTR(" - SDP Service Search Attribute Response 2: "), 0x80);
                                }

                                for (uint8_t i = 0; i < (l2capinbuf[2] + 4); i++) {
                                        D_PrintHex<uint8_t > (l2capinbuf[i], 0x80);
                                        Notify(PSTR(" "), 0x80);
                                }


                        }
                        else {
                                Notify(PSTR("\r\nUnknown PDU: "), 0x80);
                                D_PrintHex<uint8_t > (l2capinbuf[8], 0x80);
                        }
                }
                else {
                        Notify(PSTR("\r\nUnsupported L2CAP Data - Channel ID: "), 0x80);
                        D_PrintHex<uint8_t > (l2capinbuf[7], 0x80);
                        Notify(PSTR(" "), 0x80);
                        D_PrintHex<uint8_t > (l2capinbuf[6], 0x80);


                        Notify(PSTR("\r\n"), 0x80);

                        for (uint8_t i = 0; i < BULK_MAXPKTSIZE; i++) {
                                D_PrintHex<uint8_t > (l2capinbuf[i], 0x80);
                                Notify(PSTR(" "), 0x80);
                        }
                }

                SDP_task();
        }else{
          //ACL handle is not OK
          Notify(PSTR("\r\nACL handle is NOT OK: "), 0x80);
          D_PrintHex<uint8_t > (l2capinbuf[0], 0x80);
          Notify(PSTR(" "), 0x80);
          D_PrintHex<uint8_t > (l2capinbuf[1], 0x80);
          Notify(PSTR(" "), 0x80);
        }
}

void SPPi::Run() {

        if (pBtd->pairWithOtherDevice){
                if (l2cap_sdp_state == L2CAP_SDP_WAIT)
                {
                        if (pBtd->connectToOtherDevice && !pBtd->l2capConnectionClaimed && !connected) {
                                pBtd->l2capConnectionClaimed = true;
                                Notify(PSTR("\r\nSDP Connection Request"), 0x80);
                                hci_handle = pBtd->hci_handle; // Store the HCI Handle for the connection
                                l2cap_event_flag = 0; // Reset flags
                                identifier = 0;

                                pBtd->l2cap_connection_request(hci_handle, identifier, sdp_scid, SDP_PSM);
                                l2cap_sdp_state = L2CAP_SDP_REQUEST;
                        }
                }
        }
}

void SPPi::SDP_task() {
        switch (l2cap_sdp_state) {

                case L2CAP_SDP_REQUEST:
                        if (l2cap_connection_response_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_CONNECTION_SDP_RESPONSE; // Clear flag
                                identifier++;
                                l2cap_sdp_state = L2CAP_SDP_RESPONSE;
                        }
                        break;

                case L2CAP_SDP_RESPONSE:
                        if (l2cap_config_request_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_CONFIG_SDP_REQUEST; // Clear flag
                                pBtd->l2cap_config_response(hci_handle, identifier, sdp_dcid);
                                Notify(PSTR("\r\nSDP Configuration Response Sent"), 0x80);
                                l2cap_sdp_state = L2CAP_SDP_CONFIG_REQUEST;
                                identifier++;
                                delay(1);
                                Notify(PSTR("\r\nSDP Configuration Request Sent"), 0x80);
                                pBtd->l2cap_config_request(hci_handle, identifier, sdp_dcid);
                        }
                        break;


                case L2CAP_SDP_CONFIG_REQUEST:
                        if (l2cap_config_success_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_CONFIG_SDP_SUCCESS;
                                Notify(PSTR("\r\nSDP Successfully Configured"), 0x80);
                                l2cap_sdp_state = L2CAP_SDP_SERVICE_SEARCH_ATTR1;
                                identifier++;
                                SDP_Service_Search_Attr(0, identifier);
                                Notify(PSTR("\r\nSDP Service Search Attribute Request 1 Sent"), 0x80);
                        }
                        break;

                case L2CAP_SDP_SERVICE_SEARCH_ATTR1:
                        if (l2cap_service_search_attr_response_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_SERVICE_SEARCH_ATTR_RESPONSE;

                                l2cap_sdp_state = L2CAP_SDP_SERVICE_SEARCH_ATTR2;
                                identifier++;
                                if (remainingBytes)
                                {
                                    SDP_Service_Search_Attr(0, identifier, remainingBytes);
                                    Notify(PSTR("\r\nSDP Service Search Attribute Request 2 Sent"), 0x80);
                                }else 
                                    l2cap_sdp_state = L2CAP_SDP_DONE;
                        }
                        break;

                case L2CAP_SDP_SERVICE_SEARCH_ATTR2:
                        if (l2cap_service_search_attr_response_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_SERVICE_SEARCH_ATTR_RESPONSE;




                                l2cap_sdp_state = L2CAP_SDP_DONE;
                                identifier++;
                                //pBtd->l2cap_disconnection_request(hci_handle, uint8_t rxid, 
                                //        uint8_t* dcid, uint8_t* scid);
                        }
                        break;

                case L2CAP_SDP_DONE:
                        if (l2cap_disconnect_request_sdp_flag) {
                                l2cap_event_flag &= ~L2CAP_FLAG_DISCONNECT_SDP_REQUEST; // Clear flag
                                SDPConnected = false;
                                Notify(PSTR("\r\nDisconnected SDP Channel"), 0x80);
                                pBtd->l2cap_disconnection_response(hci_handle, identifier, sdp_scid, sdp_dcid);
                                l2cap_sdp_state = L2CAP_SDP_WAIT;
                        } else if (l2cap_connection_request_sdp_flag)
                                l2cap_rfcomm_state = L2CAP_SDP_WAIT;
                        break;
                case L2CAP_DISCONNECT_RESPONSE: // This is for both disconnection response from the RFCOMM and SDP channel if they were connected
                        if (l2cap_disconnect_response_flag) {
                                Notify(PSTR("\r\nDisconnected L2CAP Connection"), 0x80);
                                RFCOMMConnected = false;
                                SDPConnected = false;
                                pBtd->hci_disconnect(hci_handle);
                                hci_handle = -1; // Reset handle
                                l2cap_event_flag = 0; // Reset flags
                                l2cap_sdp_state = L2CAP_SDP_WAIT;
                                l2cap_rfcomm_state = L2CAP_RFCOMM_WAIT;
                        }
                        break;
        }
}

/************************************************************/
/*                    SDP Commands                          */

/************************************************************/
void SPPi::SDP_Command(uint8_t* data, uint8_t nbytes) { // See page 223 in the Bluetooth specs

//Notify(PSTR("\r\nSending SDP Command to dcid: "), 0x80);
//D_PrintHex<uint8_t > (sdp_dcid[0], 0x80);
//Notify(PSTR(" "), 0x80);
//D_PrintHex<uint8_t > (sdp_dcid[1], 0x80);


        pBtd->L2CAP_Command(hci_handle, data, nbytes, sdp_dcid[0], sdp_dcid[1]);
}

void SPPi::SDP_Service_Search(uint8_t transactionIDHigh, uint8_t transactionIDLow)
{
        l2capoutbuf[0] = 0x2;
        l2capoutbuf[1] = transactionIDHigh;
        l2capoutbuf[2] = transactionIDLow;
        l2capoutbuf[3] = 0x0; // Parameter Length
        l2capoutbuf[4] = 0x8; // Parameter Length

        l2capoutbuf[5] = 0x35; // Data Element Sequence, data size in next 8 bits
        l2capoutbuf[6] = 0x3; // Data size

        l2capoutbuf[7] = 0x19; // 00011 001 UUID, 2 bytes

        l2capoutbuf[8] = 0x1; 
        l2capoutbuf[9] = 0x0; // 

        l2capoutbuf[10] = 0x0;
        l2capoutbuf[11] = 0x2A;// 0x002A Maximum attribute byte count

        l2capoutbuf[12] = 0x0;// No more data

        SDP_Command(l2capoutbuf, 13);
}

void SPPi::SDP_Service_Search_Attr(uint8_t transactionIDHigh, uint8_t transactionIDLow,
        uint8_t remainingLen) {

        l2capoutbuf[0] = SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST_PDU;
        l2capoutbuf[1] = transactionIDHigh;
        l2capoutbuf[2] = transactionIDLow;

        l2capoutbuf[3] = 0x00; // Parameter Length
        if (remainingLen == 0)
            l2capoutbuf[4] = 0x0F; // Parameter Length
        else
            l2capoutbuf[4] = 0x11; // Parameter Length

        l2capoutbuf[5] = 0x35; // Data Element Sequence, data size in next 8 bits
        l2capoutbuf[6] = 0x03; // Data size

        l2capoutbuf[7] = 0x19; // 00011 001 UUID, 2 bytes
        l2capoutbuf[8] = 0x01; 
        l2capoutbuf[9] = 0x00; // 0x0100 - L2CAP_UUID
//        l2capoutbuf[8] = 0x11; 
//        l2capoutbuf[9] = 0x01; // 0x0100 - SERIAL_PORT_UUID


        l2capoutbuf[10] = 0x00;
        l2capoutbuf[11] = 0x26;// 0x0026 Maximum attribute byte count

        l2capoutbuf[12] = 0x35;// Data Element Sequence, data size in next 8 bits
        l2capoutbuf[13] = 0x05;// Data size

        l2capoutbuf[14] = 0x0A;// 00001 010 unsigned int, 4 bytes  - Attribute ID range

        l2capoutbuf[15] = 0x00;
        l2capoutbuf[16] = 0x00;// range from 0x0000...
//        l2capoutbuf[15] = 0x01;
//        l2capoutbuf[16] = 0x00;// range from 0x0000...


        l2capoutbuf[17] = 0xFF;
        l2capoutbuf[18] = 0xFF;// ... to 0xFFFF?

        if (remainingLen == 0)
        {       
            l2capoutbuf[19] = 0x00;// No more data
            SDP_Command(l2capoutbuf, 20);          
        }else{
            l2capoutbuf[19] = 0x02;
            l2capoutbuf[20] = 0x0; //will be 0 anyway
            l2capoutbuf[21] = remainingLen;
            SDP_Command(l2capoutbuf, 22);
        }
}

/************************************************************/
/*                    RFCOMM Commands                       */

/************************************************************/

void SPPi::send(void)
{

}

int SPPi::available(void)
{
    return 0;
}


int SPPi::peek(void)
{
    return 0;
}

int SPPi::read(void)
{
    return 0;
}

size_t SPPi::write(uint8_t data)
{
    return 0;
}

size_t SPPi::write(const uint8_t* data, size_t size)
{
    return 0;
}



